# -*- coding: UTF-8 -*-

import os
import time
import pickle
import argparse
import logging
import numpy as np
import pandas as pd


class BaseLoader(object):
    @staticmethod
    def parse_data_args(parser):
        parser.add_argument('--path', type=str, default='../data/',
                            help='Input data dir.')
        parser.add_argument('--dataset', type=str, default='Grocery_and_Gourmet_Food',
                            help='Choose a dataset.')
        parser.add_argument('--sep', type=str, default='\t',
                            help='sep of csv file.')
        parser.add_argument('--history_max', type=int, default=20,
                            help='Maximum length of history.')
        return parser

    def __init__(self, args):
        self.sep = args.sep
        self.prefix = args.path
        self.dataset = args.dataset
        self.history_max = args.history_max

        t0 = time.time()
        # 加载训练、测试、数据集；加载item元信息；构造包含互补和互替关系的知识图谱等
        self._load_data()
        # 提取每个样本之前的历史购买记录
        self._append_info()
        logging.info('Done! [{:<.2f} s]'.format(time.time() - t0) + os.linesep)

    def _load_data(self):
        # Loading data from "../data/", dataset = "Grocery_and_Gourmet_Food"
        logging.info('Loading data from \"{}\", dataset = \"{}\" '.format(self.prefix, self.dataset))
        self.data_df, self.item_meta_df = dict(), pd.DataFrame()
        # 读取train、dev、test数据到self.data_df中; 读取item元信息到 self.item_meta_df
        self._load_preprocessed_df()

        logging.info('Formating data type...')
        # temp = list(self.data_df.values()) + [self.item_meta_df]
        for df in list(self.data_df.values()) + [self.item_meta_df]:
            for col in df.columns:
                df[col] = df[col].apply(lambda x: eval(str(x)))

        logging.info('Constructing relation triplets...')
        self.triplet_set = set()
        # <class 'list'>: ['r_complement', 'r_substitute']
        relation_types = [r for r in self.item_meta_df.columns if r.startswith('r_')]
        # list(372090)
        heads, relations, tails = [], [], []
        for idx in range(len(self.item_meta_df)):
            head_item = self.item_meta_df.iloc[idx]['item_id']
            for r_idx, r in enumerate(relation_types):
                for tail_item in self.item_meta_df.iloc[idx][r]:
                    heads.append(head_item)
                    relations.append(r_idx + 1)
                    tails.append(tail_item)
                    self.triplet_set.add((head_item, r_idx + 1, tail_item))
        self.relation_df = pd.DataFrame()
        self.relation_df['head'] = heads
        self.relation_df['relation'] = relations
        self.relation_df['tail'] = tails

        logging.info('Counting dataset statistics...')
        self.all_df = pd.concat([df[['user_id', 'item_id', 'time']] for df in self.data_df.values()])
        # 14681  8457
        self.n_users, self.n_items = self.all_df['user_id'].max() + 1, self.all_df['item_id'].max() + 1
        # 145832
        self.n_clicks = len(self.all_df)
        self.min_time, self.max_time = self.all_df['time'].min(), self.all_df['time'].max()
        # 3, relation_types只有两种['r_complement', 'r_substitute']，可能把非这俩的算作第三种，
        # 但要注意在构造的知识图谱即self.relation_df里只有两种关系
        self.n_relations = self.relation_df['relation'].max() + 1
        # 372090
        self.n_triplets = len(self.relation_df)
        # "# users": 14681, "# items": 8457, "# clicks": 145832
        # "# relations": 3, "# triplets": 372090
        logging.info('"# users": {}, "# items": {}, "# clicks": {}'.format(self.n_users, self.n_items, self.n_clicks))
        logging.info('"# relations": {}, "# triplets": {}'.format(self.n_relations, self.n_triplets))
        # print('hi')

    def _append_info(self):
        """
        Add history info to data_df: item_his, time_his, his_length
        ! Need data_df to be sorted by time in ascending order
        :return:
        """
        logging.info('Adding history info...')
        user_his_dict = dict()  # store the already seen sequence of each user
        for key in ['train', 'dev', 'test']:
            # 注意，在原始数据的预处理文件preprocess.py里面已经把数据集按时间顺序排序了，见out_df.sort_values(by=['time',
            df = self.data_df[key]
            i_history, t_history = [], []
            for uid, iid, t in zip(df['user_id'].tolist(), df['item_id'].tolist(), df['time'].tolist()):
                if uid not in user_his_dict:
                    user_his_dict[uid] = []
                # 因为数据集是按时间排好序了的，所以这样抓取每个item的历史是没问题的
                # 可以看到因为最终i_history的长度其实是训练集的长度，即每个(user，item)样本都配备了一个该user的当前的历史购买记录，
                # 由下面的df['item_his'] = i_history也可以进一步验证。随着时间推移，显然后面的样本的历史记录会更长
                i_history.append([x[0] for x in user_his_dict[uid]])
                t_history.append([x[1] for x in user_his_dict[uid]])
                user_his_dict[uid].append((iid, t))
            # i_history: list(116489)
            # i_history[0]: <class 'list'>: []  i_history[1]: <class 'list'>: []
            # i_history[2]: <class 'list'>: []  i_history[16]: <class 'list'>: [280, 267]
            df['item_his'] = i_history
            df['time_his'] = t_history
            # self.history_max: 20
            if self.history_max > 0:
                df['item_his'] = df['item_his'].apply(lambda x: x[-self.history_max:])
                df['time_his'] = df['time_his'].apply(lambda x: x[-self.history_max:])
            df['his_length'] = df['item_his'].apply(lambda x: len(x))

        self.user_clicked_set = dict()
        for uid in user_his_dict:
            # 注意，这里存放的是融合训练和测试都有的
            self.user_clicked_set[uid] = set([x[0] for x in user_his_dict[uid]])

    def _load_preprocessed_df(self):
        # '../data/Grocery_and_Gourmet_Food/item_meta.csv'
        item_meta_path = os.path.join(self.prefix, self.dataset, 'item_meta.csv')
        if os.path.exists(item_meta_path):
            self.item_meta_df = pd.read_csv(item_meta_path, sep=self.sep)
        self.data_df['train'] = pd.read_csv(os.path.join(self.prefix, self.dataset, 'train.csv'), sep=self.sep)
        self.data_df['dev'] = pd.read_csv(os.path.join(self.prefix, self.dataset, 'dev.csv'), sep=self.sep)
        self.data_df['test'] = pd.read_csv(os.path.join(self.prefix, self.dataset, 'test.csv'), sep=self.sep)
        # len of train: 116489
        # len of dev: 14667
        # len of dev: 14676
        print('len of train:', len(self.data_df['train']))
        print('len of dev:', len(self.data_df['dev']))
        print('len of dev:', len(self.data_df['test']))
        # print('hi')


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    parser = argparse.ArgumentParser()
    parser = BaseLoader.parse_data_args(parser)
    args, extras = parser.parse_known_args()

    args.path = '../../data/'
    corpus = BaseLoader(args)

    corpus_path = os.path.join(args.path, args.dataset, 'Corpus.pkl')
    logging.info('Save corpus to {}'.format(corpus_path))
    pickle.dump(corpus, open(corpus_path, 'wb'))
