# -*- coding: UTF-8 -*-

import os
import sys
import pickle
import logging
import argparse
import numpy as np
import torch

from models import *
from helpers import *
from utils import utils


def parse_global_args(parser):
    parser.add_argument('--gpu', type=str, default='0',
                        help='Set CUDA_VISIBLE_DEVICES')
    parser.add_argument('--verbose', type=int, default=logging.INFO,
                        help='Logging Level, 0, 10, ..., 50')
    parser.add_argument('--log_file', type=str, default='',
                        help='Logging file path')
    parser.add_argument('--random_seed', type=int, default=1000,
                        help='Random seed of numpy and pytorch.')
    parser.add_argument('--load', type=int, default=0,
                        help='Whether load model and continue to train')
    parser.add_argument('--train', type=int, default=1,
                        help='To train the model or not.')
    parser.add_argument('--regenerate', type=int, default=0,
                        help='Whether to regenerate intermediate files.')
    return parser


def main():
    logging.info('-' * 45 + ' BEGIN: ' + utils.get_time() + ' ' + '-' * 45)
    logging.info(utils.format_arg_str(args))

    # Random seed
    torch.manual_seed(args.random_seed)
    torch.cuda.manual_seed(args.random_seed)
    np.random.seed(args.random_seed)

    # GPU
    # os.environ["CUDA_VISIBLE_DEVICES"] = args.gpu
    # logging.info("# cuda devices: {}".format(torch.cuda.device_count()))

    # Load data
    # 保存处理好的数据这一做法挺值得学习的
    # '../data/Grocery_and_Gourmet_Food/Corpus.pkl'     args.path: '../data/'
    corpus_path = os.path.join(args.path, args.dataset, 'Corpus.pkl')
    if not args.regenerate and os.path.exists(corpus_path):
        logging.info('Load corpus from {}'.format(corpus_path))
        corpus = pickle.load(open(corpus_path, 'rb'))
    else:
        corpus = loader_name(args)
        logging.info('Save corpus to {}'.format(corpus_path))
        pickle.dump(corpus, open(corpus_path, 'wb'))

    # Define model
    model = model_name(args, corpus)
    logging.info(model)
    model = model.double()
    model.apply(model.init_weights)
    # 注意是处理关于stage2加载stage1的预训练模型的
    model.actions_before_train()
    if torch.cuda.device_count() > 0:
        model = model.cuda()

    # Run model
    runner = runner_name(args)
    # todo 下面这行的用意应该是作者方便直接调试测试阶段的个变量的shape(比如get_feed_dict时训练和测试是不一样的)，当前我为了调试训练阶段的shape，暂时先注释掉
    # logging.info('Test Before Training: ' + runner.print_res(model, corpus))
    # 0，这个应该是应对训练突然中止然后恢复的
    if args.load > 0:
        model.load_model()
    # 1
    if args.train > 0:
        runner.train(model, corpus)
    logging.info(os.linesep + 'Test After Training: ' + runner.print_res(model, corpus))

    model.actions_after_train()
    logging.info(os.linesep + '-' * 45 + ' END: ' + utils.get_time() + ' ' + '-' * 45)


if __name__ == '__main__':
    init_parser = argparse.ArgumentParser(description='Model')
    init_parser.add_argument('--model_name', type=str, default='TRR', help='Choose a model to run.')
    # 用parse_known_args是因为这里只定义了一个参数model_name，而命令行里可能会传来很多未知参数，那些会放到init_extras里
    # Namespace(model_name='Chorus')    <class 'list'>: ['--emb_size', '64', '--margin', '1', '--lr', '5e-4', '--l2', '1e-5', '--check_epoch', '10', '--epoch', '50', '--early_stop', '0', '--batch_size', '512', '--dataset', "'Grocery_and_Gourmet_Food'", '--stage', '1']
    init_args, init_extras = init_parser.parse_known_args()
    # 注意，下面有两个重复的BaseReader，原因是第一个BaseReader是文件名，即helper文件夹下的BaseReader文件里面的BaseReader类
    # <class 'models.Chorus.Chorus'>
    model_name = eval('{0}.{0}'.format(init_args.model_name))
    # <class 'helpers.BaseLoader.BaseLoader'>
    loader_name = eval('{0}.{0}'.format(model_name.loader))
    # <class 'helpers.BaseRunner.BaseRunner'>
    runner_name = eval('{0}.{0}'.format(model_name.runner))

    # Args
    # 需要新建一个空的parser是因为下面的parse_global_args、parse_data_args等函数是直接在传入的parser上面添加参数的
    parser = argparse.ArgumentParser(description='')
    # 可见四类参数的分布
    parser = parse_global_args(parser)
    parser = loader_name.parse_data_args(parser)
    parser = runner_name.parse_runner_args(parser)
    parser = model_name.parse_model_args(parser)
    args, extras = parser.parse_known_args()

    # Logging configuration
    # <class 'list'>: ['Chorus', "'Grocery_and_Gourmet_Food'", '5']
    log_args = [init_args.model_name, args.dataset, str(args.random_seed)]
    # model_name.extra_log_args: ['margin', 'lr_scale', 'stage'], 在Chorus文件里
    for arg in ['lr', 'l2'] + model_name.extra_log_args:
        # <class 'list'>: ['Chorus', "'Grocery_and_Gourmet_Food'", '5', 'lr=0.0005', 'l2=1e-05', 'margin=1.0', 'lr_scale=0.1', 'stage=1']
        log_args.append(arg + '=' + str(eval('args.' + arg)))
    # log_args: <class 'list'>: ['Chorus', 'Grocery_and_Gourmet_Food', '5', 'lr=0.0005', 'l2=1e-05', 'margin=1.0', 'lr_scale=0.1', 'stage=1']
    # 下面的“\”是转义符，出现这个的原因是run里面的命令行中--dataset 'Grocery_and_Gourmet_Food'，应该去掉单引号
    # log_file_name: 'Chorus__Grocery_and_Gourmet_Food__5__lr=0.0005__l2=1e-05__margin=1.0__lr_scale=0.1__stage=1'
    log_file_name = '__'.join(log_args).replace(' ', '__')
    if args.log_file == '':
        # '../log/Chorus__Grocery_and_Gourmet_Food__5__lr=0.001__l2=0__margin=1__lr_scale=0.1__stage=2.txt'
        args.log_file = '../log/{}.txt'.format(log_file_name)
    if args.model_path == '':
        # '../model/Chorus/Chorus__Grocery_and_Gourmet_Food__5__lr=0.001__l2=0__margin=1__lr_scale=0.1__stage=2.pt'
        args.model_path = '../model/{}/{}.pt'.format(init_args.model_name, log_file_name)

    utils.check_dir(args.log_file)
    logging.basicConfig(filename=args.log_file, level=args.verbose)
    logging.getLogger().addHandler(logging.StreamHandler(sys.stdout))
    logging.info(init_args)

    main()
