# -*- coding: UTF-8 -*-

import torch
import numpy as np

from utils import utils
from models.BaseModel import BaseModel


class SLRC(BaseModel):
    @staticmethod
    def parse_model_args(parser, model_name='SLRC'):
        parser.add_argument('--emb_size', type=int, default=64,
                            help='Size of embedding vectors.')
        # 前三个数字显然是一天的秒数，最后的100感觉跟测试集中的100个item有关，可能跟假设买一件东西是一天为量级的有关
        parser.add_argument('--time_scalar', type=int, default=60 * 60 * 24 * 100,
                            help='Time scalar for time intervals.')
        parser.add_argument('--his_len', type=int, default=7,
                            help='Size of embedding vectors.')
        return BaseModel.parse_model_args(parser, model_name)

    def __init__(self, args, corpus):
        self.emb_size = args.emb_size
        self.time_scalar = args.time_scalar
        self.his_len = args.his_len
        self.user_num = corpus.n_users
        self.item_num = corpus.n_items
        self.relation_num = corpus.n_relations
        self.category_num = corpus.item_meta_df['category'].max() + 1
        self.item2cate = dict(zip(corpus.item_meta_df['item_id'].values, corpus.item_meta_df['category'].values))
        BaseModel.__init__(self, model_path=args.model_path)

    def _define_params(self):
        self.u_embeddings = torch.nn.Embedding(self.user_num, self.emb_size)
        self.i_embeddings = torch.nn.Embedding(self.item_num, self.emb_size)
        self.user_bias = torch.nn.Embedding(self.user_num, 1)
        self.item_bias = torch.nn.Embedding(self.item_num, 1)
        self.alphas = torch.nn.Embedding(self.category_num, self.relation_num)
        self.pis = torch.nn.Embedding(self.category_num, self.relation_num)
        self.betas = torch.nn.Embedding(self.category_num, self.relation_num)
        self.sigmas = torch.nn.Embedding(self.category_num, self.relation_num)
        self.mus = torch.nn.Embedding(self.category_num, self.relation_num)
        self.embeddings = ['u_embeddings', 'i_embeddings', 'user_bias', 'item_bias',
                           'alphas', 'pis', 'betas', 'sigmas', 'mus']

    def forward(self, feed_dict):
        self.check_list, self.embedding_l2 = [], []
        u_ids = feed_dict['user_id']  # [batch_size]
        i_ids = feed_dict['item_id']  # [batch_size, -1]
        c_ids = feed_dict['category_id']  # [batch_size, -1]
        r_interval = feed_dict['relational_interval']  # [batch_size, -1, relation_num]
        batch_size = feed_dict['batch_size']

        # Excitation
        alphas, pis = self.alphas(c_ids), self.pis(c_ids) + 0.5
        betas, mus = self.betas(c_ids) + 1., self.mus(c_ids) + 1.
        sigmas = (self.sigmas(c_ids) + 1.).clamp(min=1e-10, max=10)
        mask = (r_interval >= 0).double()
        delta_t = r_interval * mask
        norm_dist = torch.distributions.normal.Normal(mus, sigmas)
        exp_dist = torch.distributions.exponential.Exponential(betas)
        decay = pis * exp_dist.log_prob(delta_t).exp() + (1. - pis) * norm_dist.log_prob(delta_t).exp()
        excitation = (alphas * decay * mask).sum(-1)  # [batch_size, -1]

        # Base Intensity (CF)
        u_bias = self.user_bias(u_ids)
        i_bias = self.item_bias(i_ids).squeeze(-1)
        cf_u_vectors = self.u_embeddings(u_ids)
        cf_i_vectors = self.i_embeddings(i_ids)
        self.embedding_l2.extend([cf_u_vectors, cf_i_vectors.view(-1, self.emb_size)])
        base_intensity = (cf_u_vectors[:, None, :] * cf_i_vectors).sum(-1)
        base_intensity = base_intensity + u_bias + i_bias

        prediction = base_intensity + excitation

        out_dict = {'prediction': prediction.view(batch_size, -1), 'check': self.check_list}
        return out_dict

    # data是BaseRunner下predict函数里的corpus.data_df[phase]
    def get_feed_dict(self, corpus, data, batch_start, batch_size, phase):
        batch_end = min(len(data), batch_start + batch_size)
        real_batch_size = batch_end - batch_start
        # <class 'tuple'>: (256,)
        user_ids = data['user_id'][batch_start: batch_start + real_batch_size].values
        # <class 'tuple'>: (256,)
        item_ids = data['item_id'][batch_start: batch_start + real_batch_size].values
        # list(256) history_items[0]: list()
        history_items = data['item_his'][batch_start: batch_start + real_batch_size].values
        history_times = data['time_his'][batch_start: batch_start + real_batch_size].values
        times = data['time'][batch_start: batch_start + real_batch_size].values

        history_len = self.his_len
        user_his = np.zeros((real_batch_size, history_len), dtype=np.int64)
        for batch_ind in range(real_batch_size):
            his_len = len(history_items[batch_ind][-history_len:])
            if his_len == 0:
                continue
            else:
                user_his[batch_ind][-his_len:] = history_items[batch_ind][-history_len:]

        # 训练阶段是一个负样本，训练阶段是99个
        # 注意，负样本是针对user的，即该user从没点击过的里面随机选的
        # list(256) neg_items[0]: <class 'tuple'>: (256, 1)
        neg_items = self.get_neg_items(corpus, data, batch_start, real_batch_size, phase)
        # 训练: <class 'tuple'>: (256, 2)  测试: <class 'tuple'>: (256, 100)
        item_ids = np.concatenate([np.expand_dims(item_ids, -1), neg_items], axis=1)

        # Find information related to the target item:
        # - category id
        # - time intervals w.r.t. recent relational interactions (-1 if not existing)
        # 训练: <class 'tuple'>: (256, 2)  测试: <class 'tuple'>: (256, 100)
        category_ids = np.array([[self.item2cate[x] for x in candidate_lst] for candidate_lst in item_ids])
        relational_intervals = list()
        relation_item_inds = list()
        # 如图3所示，注意检查每一个关系的存在
        for r_idx in range(0, self.relation_num):
            # <class 'tuple'>: (256, 100)
            intervals = np.ones_like(item_ids) * -1.
            relation_item_ind = np.zeros_like(item_ids)
            # 这一层循环是针对batch维的，所以重点可以不看这一层，而看里面那层，所以把这一层固定住
            # candidate_lst: <class 'tuple'>: (100,)
            for i, candidate_lst in enumerate(item_ids):
                for j, target_item in enumerate(candidate_lst):
                    for k in range(len(history_items[i]))[::-1]:
                        # 在固定item下，即固定i下，下面展示的就是对于目标item，该user过往浏览item与该目标item的关系，是互补的还是替换的
                        # 注意，在corpus.triplet_set是只记录了互补或互替关系的，不包括索引为0的那种关系
                        if (history_items[i][k], r_idx, target_item) in corpus.triplet_set:
                            # 不论互补还是互替关系存在，记录下该历史item与当前目标item的时间距离
                            intervals[i][j] = times[i] - history_times[i][k]
                            relation_item_ind[i][j] = history_items[i][k]
                            # break的原因大概是每个目标item只在历史序列里从后往前依次寻找某个关系(互补或互替)，找到一个就停止，
                            # 比如某轮循环里找互补关系，找到最近的那个就停止，不会继续再往前找该同种关系，可能是因为最近的那个是主要影响因素，
                            # 每个关系的寻找逻辑都同此。见论文式9上面的表述，说的就是这个
                            break
            relational_intervals.append(intervals)
            relation_item_inds.append(relation_item_ind)
        # 训练: <class 'tuple'>: (256, 2, 3)  测试: <class 'tuple'>: (256, 100, 3)
        relational_intervals = np.stack(relational_intervals, axis=2) / self.time_scalar
        relation_item_inds = np.stack(relation_item_inds, axis=2)

        feed_dict = {
            'user_id': utils.numpy_to_torch(user_ids),  # [batch_size]
            'item_id': utils.numpy_to_torch(item_ids),  # [batch_size, -1]
            'user_his': utils.numpy_to_torch(user_his),
            'category_id': utils.numpy_to_torch(category_ids),  # [batch_size, -1]
            'relational_interval': utils.numpy_to_torch(relational_intervals),  # [batch_size, -1, relation_num]
            'relation_item_inds': utils.numpy_to_torch(relation_item_inds),
            'batch_size': real_batch_size
        }
        return feed_dict
